@@ -15,8 +15,10 @@ import functools
 import hashlib
 import json
 import logging
+import os
 import platform
 import socket
+import sys
 import time
 import uuid
 
@@ -72,7 +74,7 @@ class _JSONEncoder(json.JSONEncoder):
 
 
 class _StringFormatter(object):
-    """A String formatter that fetches values on demand"""
+    """A String formatter that fetches values on demand."""
 
     def __init__(self, session, auth):
         self.session = session
@@ -93,6 +95,75 @@ class _StringFormatter(object):
         return value
 
 
+def _determine_calling_package():
+    """Walk the call frames trying to identify what is using this module."""
+    # Create a lookup table mapping file name to module name. The ``inspect``
+    # module does this but is far less efficient. Same story with the
+    # frame walking below.  One could use ``inspect.stack()`` but it
+    # has far more overhead.
+    mod_lookup = dict((m.__file__, n) for n, m in sys.modules.items()
+                      if hasattr(m, '__file__'))
+
+    # NOTE(shaleh): these are not useful because they hide the real
+    # user of the code. debtcollector did not import keystoneauth but
+    # it will show up in the call stack. Similarly we do not want to
+    # report ourselves or keystone client as the user agent. The real
+    # user is the code importing them.
+    ignored = ('debtcollector', 'keystoneauth1', 'keystoneclient')
+
+    i = 0
+    while True:
+        i += 1
+
+        try:
+            # NOTE(shaleh): this is safe in CPython but could break in
+            # other implementations of Python. Yes, the `inspect`
+            # module could be used instead. But it does a lot more
+            # work so it has worse performance.
+            f = sys._getframe(i)
+            try:
+                name = mod_lookup[f.f_code.co_filename]
+                # finds the full name module.foo.bar but all we need
+                # is the module name.
+                name, _, _ = name.partition('.')
+                if name not in ignored:
+                    return name
+            except KeyError:
+                pass  # builtin or the like
+        except ValueError:
+            # hit the bottom of the frame stack
+            break
+
+    return ''
+
+
+def _determine_user_agent():
+    """Attempt to programatically generate a user agent string.
+
+    First, look at the name of the process. Return this unless it is in
+    the `ignored` list.  Otherwise, look at the function call stack and
+    try to find the name of the code that invoked this module.
+    """
+    # NOTE(shaleh): mod_wsgi is not any more useful than just
+    # reporting "keystoneauth". Ignore it and perform the package name
+    # heuristic.
+    ignored = ('mod_wsgi', )
+
+    try:
+        name = sys.argv[0]
+    except IndexError:
+        # sys.argv is empty, usually the Python interpreter prevents this.
+        return ''
+
+    if not name:
+        return ''
+
+    name = os.path.basename(name)
+    if name in ignored:
+        name = _determine_calling_package()
+    return name
+
+
 class Session(object):
     """Maintains client communication state and common functionality.
 
@@ -132,18 +203,32 @@ class Session(object):
                               can be followed by a request. Either an integer
                               for a specific count or True/False for
                               forever/never. (optional, default to 30)
+    :param dict additional_headers: Additional headers that should be attached
+                                    to every request passing through the
+                                    session. Headers of the same name specified
+                                    per request will take priority.
+    :param str app_name: The name of the application that is creating the
+                         session. This will be used to create the user_agent.
+    :param str app_version: The version of the application creating the
+                            session. This will be used to create the
+                            user_agent.
+    :param list additional_user_agent: A list of tuple of name, version that
+                                       will be added to the user agent. This
+                                       can be used by libraries that are part
+                                       of the communication process.
     """
 
     user_agent = None
 
-    _REDIRECT_STATUSES = (301, 302, 303, 305, 307)
+    _REDIRECT_STATUSES = (301, 302, 303, 305, 307, 308)
 
     _DEFAULT_REDIRECT_LIMIT = 30
 
     @positional(2)
     def __init__(self, auth=None, session=None, original_ip=None, verify=True,
                  cert=None, timeout=None, user_agent=None,
-                 redirect=_DEFAULT_REDIRECT_LIMIT):
+                 redirect=_DEFAULT_REDIRECT_LIMIT, additional_headers=None,
+                 app_name=None, app_version=None, additional_user_agent=None):
 
         self.auth = auth
         self.session = _construct_session(session)
@@ -152,16 +237,16 @@ class Session(object):
         self.cert = cert
         self.timeout = None
         self.redirect = redirect
+        self.additional_headers = additional_headers or {}
+        self.app_name = app_name
+        self.app_version = app_version
+        self.additional_user_agent = additional_user_agent or []
+        self._determined_user_agent = None
 
         if timeout is not None:
             self.timeout = float(timeout)
 
-        # don't override the class variable if none provided
         if user_agent is not None:
-            # Per RFC 7231 Section 5.5.3, identifiers in a user-agent
-            # should be ordered by decreasing significance.
-            # If a user sets their product, we prepend it to the KSA
-            # version, requests version, and then the Python version.
             self.user_agent = "%s %s" % (user_agent, DEFAULT_USER_AGENT)
 
         self._json = _JSONEncoder()
@@ -198,7 +283,7 @@ class Session(object):
 
     @staticmethod
     def _process_header(header):
-        """Redacts the secure headers to be logged."""
+        """Redact the secure headers to be logged."""
         secure_headers = ('authorization', 'x-auth-token',
                           'x-subject-token',)
         if header[0].lower() in secure_headers:
@@ -210,7 +295,8 @@ class Session(object):
 
     @positional()
     def _http_log_request(self, url, method=None, data=None,
-                          json=None, headers=None, logger=_logger):
+                          json=None, headers=None, query_params=None,
+                          logger=_logger):
         if not logger.isEnabledFor(logging.DEBUG):
             # NOTE(morganfainberg): This whole debug section is expensive,
             # there is no need to do the work if we're not going to emit a
@@ -229,7 +315,15 @@ class Session(object):
         if method:
             string_parts.extend(['-X', method])
 
-        string_parts.append(url)
+        if query_params:
+            # Don't check against `is not None` as this can be
+            # an empty dictionary, which we shouldn't bother logging.
+            url = url + '?' + urllib.parse.urlencode(query_params)
+            # URLs with query strings need to be wrapped in quotes in order
+            # for the CURL command to run properly.
+            string_parts.append('"%s"' % url)
+        else:
+            string_parts.append(url)
 
         if headers:
             for header in six.iteritems(headers):
@@ -282,7 +376,7 @@ class Session(object):
                 endpoint_filter=None, auth=None, requests_auth=None,
                 raise_exc=True, allow_reauth=True, log=True,
                 endpoint_override=None, connect_retries=0, logger=_logger,
-                **kwargs):
+                allow={}, client_name=None, client_version=None, **kwargs):
         """Send an HTTP request with the specified characteristics.
 
         Wrapper around `requests.Session.request` to handle tasks such as
@@ -348,6 +442,8 @@ class Session(object):
                        If not provided the keystoneauth1.session default
                        logger will be used.
         :type logger: logging.Logger
+        :param dict allow: Extra filters to pass when discovering API
+                           versions. (optional)
         :param kwargs: any other parameter that can be passed to
                        :meth:`requests.Session.request` (such as `headers`).
                        Except:
@@ -389,7 +485,8 @@ class Session(object):
             if endpoint_override:
                 base_url = endpoint_override % _StringFormatter(self, auth)
             elif endpoint_filter:
-                base_url = self.get_endpoint(auth, **endpoint_filter)
+                base_url = self.get_endpoint(auth, allow=allow,
+                                             **endpoint_filter)
 
             if not base_url:
                 raise exceptions.EndpointNotFound()
@@ -407,25 +504,67 @@ class Session(object):
         elif self.user_agent:
             user_agent = headers.setdefault('User-Agent', self.user_agent)
         else:
-            user_agent = headers.setdefault('User-Agent', DEFAULT_USER_AGENT)
+            # Per RFC 7231 Section 5.5.3, identifiers in a user-agent should be
+            # ordered by decreasing significance.  If a user sets their product
+            # that value will be used. Otherwise we attempt to derive a useful
+            # product value. The value will be prepended it to the KSA version,
+            # requests version, and then the Python version.
+
+            agent = []
+
+            if self.app_name and self.app_version:
+                agent.append('%s/%s' % (self.app_name, self.app_version))
+            elif self.app_name:
+                agent.append(self.app_name)
+
+            for additional in self.additional_user_agent:
+                agent.append('%s/%s' % additional)
+
+            if client_name and client_version:
+                agent.append('%s/%s' % (client_name, client_version))
+            elif client_name:
+                agent.append(client_name)
+
+            if not agent:
+                # NOTE(jamielennox): determine_user_agent will return an empty
+                # string on failure so checking for None will ensure it is only
+                # called once even on failure.
+                if self._determined_user_agent is None:
+                    self._determined_user_agent = _determine_user_agent()
+
+                if self._determined_user_agent:
+                    agent.append(self._determined_user_agent)
+
+            agent.append(DEFAULT_USER_AGENT)
+            user_agent = headers.setdefault('User-Agent', ' '.join(agent))
 
         if self.original_ip:
             headers.setdefault('Forwarded',
                                'for=%s;by=%s' % (self.original_ip, user_agent))
 
         if json is not None:
-            headers['Content-Type'] = 'application/json'
+            headers.setdefault('Content-Type', 'application/json')
             kwargs['data'] = self._json.encode(json)
 
+        for k, v in self.additional_headers.items():
+            headers.setdefault(k, v)
+
         kwargs.setdefault('verify', self.verify)
 
         if requests_auth:
             kwargs['auth'] = requests_auth
 
+        # Query parameters that are included in the url string will
+        # be logged properly, but those sent in the `params` parameter
+        # (which the requests library handles) need to be explicitly
+        # picked out so they can be included in the URL that gets loggged.
+        query_params = kwargs.get('params', dict())
+
         if log:
             self._http_log_request(url, method=method,
                                    data=kwargs.get('data'),
                                    headers=headers,
+                                   query_params=query_params,
                                    logger=logger)
 
         # Force disable requests redirect handling. We will manage this below.
@@ -492,8 +631,14 @@ class Session(object):
             except requests.exceptions.Timeout:
                 msg = 'Request to %s timed out' % url
                 raise exceptions.ConnectTimeout(msg)
-            except requests.exceptions.ConnectionError:
-                msg = 'Unable to establish connection to %s' % url
+            except requests.exceptions.ConnectionError as e:
+                # NOTE(sdague): urllib3/requests connection error is a
+                # translation of SocketError. However, SocketError
+                # happens for many different reasons, and that low
+                # level message is often really important in figuring
+                # out the difference between network misconfigurations
+                # and firewall blocking.
+                msg = 'Unable to establish connection to %s: %s' % (url, e)
                 raise exceptions.ConnectFailure(msg)
             except requests.exceptions.RequestException as e:
                 msg = 'Unexpected exception for %(url)s: %(error)s' % {
@@ -695,8 +840,7 @@ class Session(object):
         :returns: Authentication headers or None for failure.
         :rtype: :class:`dict`
         """
-        msg = 'An auth plugin is required to fetch connection params'
-        auth = self._auth_required(auth, msg)
+        auth = self._auth_required(auth, 'fetch connection params')
         params = auth.get_connection_params(self, **kwargs)
 
         # NOTE(jamielennox): There needs to be some consensus on what
@@ -774,6 +918,7 @@ class TCPKeepAliveAdapter(requests.adapters.HTTPAdapter):
     disables Nagle's Algorithm. See also:
     http://blogs.msdn.com/b/windowsazurestorage/archive/2010/06/25/nagle-s-algorithm-is-not-friendly-towards-small-requests.aspx
     """
+
     def init_poolmanager(self, *args, **kwargs):
         if 'socket_options' not in kwargs and REQUESTS_VERSION >= (2, 4, 1):
             socket_options = [
@@ -791,7 +936,9 @@ class TCPKeepAliveAdapter(requests.adapters.HTTPAdapter):
                     (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 60)
                 ]
 
-            if hasattr(socket, 'TCP_KEEPCNT'):
+            # Windows subsystem for Linux does not support this feature
+            if (hasattr(socket, 'TCP_KEEPCNT') and
+                    not utils.is_windows_linux_subsystem):
                 socket_options += [
                     # Set the maximum number of keep-alive probes
                     (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 4),
